
(set_backtrace t)

(require 'multisyn/multisyn_pauses)

(defvar closure_regex ".*_cl")
(defvar sp_threshold 0.060) ; was 0.050 for utts 5, was 0.02 for 5b
(defvar sp_closure_threshold 0.100)

(defvar multisyn_lib_dir (path-append libdir "multisyn/"))
(if (not (member_string multisyn_lib_dir libdir))
    (set! load-path (cons multisyn_lib_dir load-path)))

(defvar msb_ofd nil)
(defvar msb_lab nil)
(defvar msb_name nil)

(set! debug nil)

;;;; Function to check an utts.data file for out of lexicon words
(define (check_script infile lexicon)
"(check_script infile lexicon)
Check the script for out of lexicon words."
  (let (fd p (c 0))
    (setup_phoneset_and_lexicon lexicon)

;; Open input file
  (set! p (load infile t))

  (mapcar
     (lambda (l)
       (set! c (+ c (check_script_entry (cadr l)))))
     p)
  (format t "Found %d out of lexicon words.\n" c)
t))

;; Checks a text string for out of lexicon words
(define (check_script_entry text)
  "(check_script_entry entry)
Check the words to see if they are in the lexicon."
  (let ( (c 0)
	 (utt1 (utt.synth_toSegment_text (eval (list 'Utterance 'Text text)))))
    (mapcar
     (lambda (word)
       (if (not (lex.lookup_all (item.feat word 'name)))
	   (begin
	     (format t "\"%s\" not in lexicon.  lts give %l\n" (item.feat word 'name) (lex.lookup (item.feat word 'name)))
	     (set! c (+ c 1)))))
     (utt.relation.items utt1 'Word))
c))

(define (utt.mark_outoflex_words utt)
"(utt.mark_outoflex_words utt)
Mark each segment in out of lexicon words in an utterance as bad."
(let ()
  (mapcar
     (lambda (word)
       (if (not (lex.lookup_all (item.feat word 'name)))
           (mapcar
            (lambda (seg)
              (item.set_feat seg "bad_lex" "lts"))
            (item.leafs word 'Sylstructure))))
     (utt.relation.items utt 'Word))))

;;;; Make masterlabel file and train.scp file for htk labelling
(define (make_initial_phone_labs infile outfile lexicon)
"(make_initial_phone_labs infile outfile lexicon)

 infile        \"utts.data\"
 outfile       \"utts.mlf\"
 lexicon       'unilex-rpx

 Make the master label file for HTK labelling.

 Note: An older version of this used to also produce a list
 of mfcc files (train.scp). Use bin/make_mfcc_list now.

 If the symbol lts-trace is bound, an LTS trace is written to
 the files $utt.lts containing the following info:
 1 word name
 2 token it comes from
 3 utterance name
 4 'LEX' or 'LTS' depending on whether it is in the lexicon
 5- phone transcription
"
  (let (fd p ofd)

  (setup_phoneset_and_lexicon lexicon)

;; Open output file.  
  (set! ofd (fopen outfile "w"))
;; Open input file
  (set! p (load infile t))
  (mapcar
     (lambda (l)
       (format t "%s\n" (car l))
       (format t "%s\n" (cadr l))
       (cond
	((eq? (cadr l) 'apml)
	   (do_synth_for_htk_apml (car l)  ofd))
	((eq? (cadr l) 'xxml)
	   (do_synth_for_htk_singing_xxml (car l)  ofd))
	(t
	   (do_synth_for_htk_text (car l) (cadr l) ofd)))
       t)
     p)
  (fclose ofd)
t))


(define (remake_initial_phone_labs infile outfile lexicon)
"(remake_initial_phone_labs infile outfile)

 infile        \"utts.data\"
 outfile       \"utts.mlf\"
 lexicon       'unilex-rpx

 Remake the master label file for HTK labelling from the
./utt/*.utt files.  The names (and only the names!) are
taken from infile.

This might be useful for two reasons:
* If an alignment.3.mlf is needed as initial alignment.0.mlf for
  yet another alignment, but became dated because of
  `do_final_alignment some_utts' (which remakes the .lab files, but
  leaves alignment.3.mlf untouched), build_utts makes .utt files 
  from .lab files (and other streams), and remake_initial_phone_labs 
  then makes a .mlf from the .utt files.
* build_utts suppresses /sil/ coming from .lab files where festival
  does not expect them (they may come from post-processing the .lab
  file with e.g. label_fix_plosives).  Also, via align_utt_internal,
  it splits the actual /sil/ into /sil sil .../ with a max duration
  of 150ms each.  Repeating the alignment with these labels may be
  better than with trimmed .wav files.
"
  (let (fd p ofd)

  (setup_phoneset_and_lexicon lexicon)

  (set! ofd (fopen outfile "w"))
  (set! p (load infile t))
  (mapcar
     (lambda (l)
       (format t "%s \n" (car l))
       (load_synth_for_htk_text (car l) ofd)
     )
     p)
  (fclose ofd)
t))

(define (make_initial_utt_files infile lexicon)
"(make_initial_utt_files infile lexicon)

 infile        \"utts.data\"
 lexicon       'unilex-rpx

 Make the initial utt files for text selection.
 Use script phone_string_reformat then to make phone 
 strings from them, in which stress and boundary
 markers are attached to the phone symbols.

Output is written to fthe files utt/NAME.utt. 
Directory ./utt must exist.

"
  (setup_phoneset_and_lexicon lexicon)

  (set! p (load infile t))
  (mapcar
     (lambda (l)
        (format t "%s\n" (car l))
        (do_synth_for_text_selection (car l)  (cadr l)))
     p)
t)

;; synth to segments and save as mlf.
(define (do_synth_for_htk_text name text ofd)
  "(do_synth_for_htk_text name text ofd)
Synthesize given text and save waveform and labels for prompts."
  (let ((utt1 (utt.synth_toSegment_text (eval (list 'Utterance 'Text text)))))
    (utt.save.mlf utt1 (format nil "%s.lab" name) ofd)

    (if(symbol-bound? 'lts-trace)
       (begin
          (set! ltsfd (fopen (format nil "%s.lts" name) "w"))
          (mapcar
             (lambda(word)
                (format ltsfd "%s\t%s\t%s\t"
                   (item.name word)
                   (item.name(item.parent(item.relation word 'Token)))
                   name
                )
                (if(lex.lookup_all (item.name word))
                   (format ltsfd "LEX\t")
                   (format ltsfd "LTS\t"))
                (mapcar
                   (lambda(seg)
                      (format ltsfd "%s " (item.name seg)))
                   (item.leafs(item.relation word 'SylStructure)))
                (format ltsfd "\n")
             )
             (utt.relation.items utt1 'Word))
             (fclose ltsfd)))))

;; synth to segments and save as mlf.
(define (do_synth_for_htk_apml name ofd)
  "(do_synth_for_htk_apml name ofd)
Synthesize given apml and save waveform and labels for prompts."
  (let ((utt1 (Utterance Tokens nil)))
    (utt.load utt1 (format nil "apml/%s.apml" name))
    (utt.synth_toSegment_tokens utt1)
    (utt.save.mlf utt1 (format nil "%s.lab" name) ofd)))

;; synth to segments and save as .utt
(define (do_synth_for_text_selection name text)
  "(do_synth_for_text_selection name text ofname)
Synthesize given text to segments and save utterance in file utt/$name.utt"
  (let ((utt1 (utt.synth_toSegment_text (eval (list 'Utterance 'Text text)))))
    (utt.save utt1 (format nil "utt/%s.utt" name) nil)))

;; synth to segments and save as mlf.
(define (do_synth_for_htk_singing_xxml name ofd)
  "(do_synth_for_htk_singing_xxml name ofd)
Synthesize given xxml and save waveform and labels for prompts."
  (let ((old_tts_hooks tts_hooks))
    (set! tts_hooks (list utt.synth_toSegment_tokens xxml_mlf_save_wrapper))
    (set! msb_lab (format nil "%s.lab" name))
    (set! msb_ofd ofd)
    (tts (format nil "xml/%s.xml" name) 'singing)
    (set! tts_hooks old_tts_hooks)))

(define (xxml_mlf_save_wrapper utt)
  (format t "In mlf wrapper\n")
  (utt.save.mlf utt msb_lab msb_ofd))


;; Do the linguistic side of synthesis.
(define (utt.synth_toSegment_text utt)
  (Initialize utt)
  (Text utt)
  (Token_POS utt)    ;; when utt.synth is called
  (Token utt)
  (POS utt)
  (Phrasify utt)
  (Word utt)
  (Pauses utt)
  (Intonation utt)
  (PostLex utt))

(define (utt.synth_toWord_text utt)
  (Initialize utt)
  (Text utt)
  (Token_POS utt)    ;; when utt.synth is called
  (Token utt)
  (POS utt)
  (Phrasify utt))

(define (utt.synth_toWord_post_text utt)
  (remove_punc_words utt)
  (Intonation utt)
  ;(PostLex utt)
)


(define (utt.synth_toSegment_tokens utt)
  (Token_POS utt)    ;; when utt.synth is called
  (Token utt)
  (POS utt)
  (Phrasify utt)
  (Word utt)
  (Pauses utt)
  (Intonation utt)
  (PostLex utt))


;; Output phones in mlf format, with a few additions.
(define (utt.save.mlf utt file ofd)
  "(utt.save.mlf utt file)
Save segments and other info for aligning."
  (format ofd "#!MLF!#\n")
  (format ofd "\"*/%s\"\n" file)
  (mapcar
   (lambda (seg)
     (if (or (string-equal (item.feat seg 'ph_ctype) "s") ; add stop and affricate closures
	     (string-equal (item.feat seg 'ph_ctype) "a"))
	 (format ofd "%s_cl\n" (make_safe_name seg)))
     (if (or (string-equal (item.feat seg 'name) "#")    ; substitute # with sil - hack!
	     (string-equal (item.feat seg 'name) "pau"))
;	 [format ofd "%s\n%s\n" "sil" "sil"] ; adding 2 sils works better with big pauses.
	 (format ofd "%s\n" "sil")
	 (format ofd "%s\n" (make_safe_name seg)))
     (if (seg_word_final seg)                        ; add sp at word boundaries
	 (if (and (item.next seg)
		  (not (or (string-equal (item.feat (item.next seg) 'ph_ctype) "s") ; stop and affricate
			   (string-equal (item.feat (item.next seg) 'ph_ctype) "a"))))
	     (format ofd "sp\n"))))
   (utt.relation.items utt 'Segment))
  (format ofd ".\n")
  utt)

;;;; Substitutes phone names with ones that don't causes problems,
;;   If you change this make sure you change make_unsafe_name as well to undo the change.
(define (make_safe_name seg)
"(make_safe_name seg)
Check and replace segment names that cause problems."
(let ((name (item.feat seg "name")))
  (cond
   ((string-equal name "?") "gstop")
   ((string-equal name "3") "three")
   ((string-equal name "5") "five")
   (t name))))

;; see above.
(define (make_unsafe_name seg)
"(make_unsafe_name seg)
Check and unreplace segment names that cause problems."
(let ((name (item.feat seg "name")))
  (cond
   ((string-equal name "gstop") "?")
   ((string-equal name "three") "3")
   ((string-equal name "five") "5")
   (t name))))

;; Is this the last segment in a word [more complicated than you would think]
(define (seg_word_final seg)
"(seg_word_final seg)
Is this segment word final?"
  (let ((this_seg_word (item.parent (item.relation.parent  seg 'SylStructure)))
	(silence (car (cadr (car (PhoneSet.description '(silences))))))
	next_seg_word)
    (if (item.next seg)
	(set! next_seg_word (item.parent (item.relation.parent (item.next seg) 'SylStructure))))
    (if (or (equal? this_seg_word next_seg_word)
	     (string-equal (item.feat seg "name") silence))
	nil
	t)))
;;;; Set up the phoneset and lexicon for a given proto-voice.
;;;; If the lexicon you want to use is not listed below, you will need to add your own section.
(define (setup_phoneset_and_lexicon lexicon)
  (cond
   ;; UNILEX SETUP
   ((string-matches lexicon "unilex-...")
    ;; Explicitly select phoneset.
    (require 'unilex_phones)
    (Parameter.set 'PhoneSet 'unilex)
    (PhoneSet.select 'unilex)
  ;; And lexicon.
    (if (not (member_string lexicon (lex.list)))
	(load (path-append lexdir "unilex/" (string-append lexicon ".scm"))))
    (lex.select lexicon)
    ;; And durations
    (load (path-append lexdir "unilex/unilexdurtreeZ.scm"))
    (set! duration_cart_tree unilex_duration_cart_tree)
    (set! duration_ph_info unilex_durs)
    ;; And post lexical rules
    (if(symbol-bound? 'postlex_rules_hooks)
        (format t "\n\tsetup_phoneset_and_lexicon: using the already defined
postlex_rules_hooks %l\n" postlex_rules_hooks)
        (begin

           (set! postlex_rules_hooks (list postlex_apos_s_check postlex_the_vs_thee))
           (format t "\n\tsetup_phoneset_and_lexicon: using default postlex_rules_hooks %l\n" postlex_rules_hooks)))
    ;; Use unilex Word module.
    (Parameter.set 'Word_Method Unilex_Word))
   ;;=======================================================================
;; Combilex SETUP
   ((string-matches lexicon "combilex-...")
    ;; Explicitly select phoneset.
    (require 'combilex_phones)
    (Parameter.set 'PhoneSet 'combilex)
    (PhoneSet.select 'combilex)
  ;; And lexicon.
    (if (not (member_string lexicon (lex.list)))
	(load (path-append lexdir "combilex/" (string-append lexicon ".scm"))))
    (lex.select lexicon)
    ;; And durations
    (load (path-append lexdir "combilex/combilexdurtreeZ.scm"))
    (set! duration_cart_tree unilex_duration_cart_tree)
    (set! duration_ph_info combilex_durs)
    ;; And post lexical rules
    (if (not (symbol-bound? 'postlex_rules_hooks))
      (if (string-matches lexicon "combilex-rpx")   
        (set! postlex_rules_hooks (list postlex_apos_s_check postlex_the_vs_thee postlex_intervoc_r))
        (set! postlex_rules_hooks (list postlex_apos_s_check postlex_the_vs_thee))))
    ;; Use unilex Word module.
    (Parameter.set 'Word_Method Unilex_Word))
   ;;=======================================================================
   ;; German SAMPA SETUP
   ((or(string-equal lexicon "sampa-german")
       (string-equal lexicon "sampa-viennese"))
    ;; Explicitly select phoneset.
    (require 'postlex_german)
    (require 'sampa_phones)
    (Parameter.set 'PhoneSet 'sampa)
    (PhoneSet.select 'sampa)
  ;; And lexicon.
    (if (not (member_string lexicon (lex.list)))
        (load (path-append lexdir "sampa/" (string-append lexicon ".scm"))))
    (lex.select lexicon)
    ;; And dumy durations
    (load (path-append lexdir "sampa/sampadurtreeZ.scm"))
    (set! duration_cart_tree sampa_duration_cart_tree)
    (set! duration_ph_info sampa_durs)
    ;; And post lexical rules
    (if(symbol-bound? 'postlex_rules_hooks)
        (format t "\n\tsetup_phoneset_and_lexicon: using the already defined postlex_rules_hooks %l\n" postlex_rules_hooks)
        (begin

           (set! postlex_rules_hooks (list postlex_ger_sz))
           (format t "\n\tsetup_phoneset_and_lexicon: using default postlex_rules_hooks %l\n" postlex_rules_hooks)))
    ;; Use sampa Word module.
    (Parameter.set 'Word_Method Classic_Word))
   ;;=======================================================================
   ;; OED setup
   ((string-equal lexicon "oed")
    (require 'oed_phones)
    (Parameter.set 'PhoneSet 'oed)
    (PhoneSet.select 'oed))
    ;; And lexicon.
    ;; And durations
    ;; And post lexical rules
    ;; Use unilex Word module.
   ((string-equal lexicon "cmu")
    ;; Explicitly select phoneset.
    (require 'radio_phones_multisyn)
    (Parameter.set 'PhoneSet 'radio_multisyn)
    (PhoneSet.select 'radio_multisyn)
  ;; And lexicon.
    (setup_cmu_lex)
    (lex.select lexicon)
    ;; And durations
    (Parameter.set 'Duration_Method 'nil)
    ;; And post lexical rules
    (set! postlex_rules_hooks (list postlex_apos_s_check postlex_the_vs_thee))
    ;; Use unilex Word module.
    (Parameter.set 'Word_Method Classic_Word))
   ((string-equal lexicon "iblex")
    ;; Explicitly select phoneset.
    (require 'ibibio_phones)
    (ibibio::select_phoneset)
  ;; And lexicon.
    (if (not (member_string lexicon (lex.list)))
	(load (path-append lexdir "iblex/" (string-append lexicon ".scm"))))
    (lex.select lexicon)
    ;; And durations
    (Parameter.set 'Duration_Method 'nil)
    ;; And post lexical rules
    (set! postlex_rules_hooks nil)
    ;; Use unilex Word module.
    (Parameter.set 'Word_Method Classic_Word))

   ))

;; Alignment stuff [From festvox limited domain code]

(define (build_pauses_utts file lexicon)
  "(build_pauses_utts file) 
For each utterances in a pause prompt file, build utterances."
  (setup_phoneset_and_lexicon lexicon)
  (let ((p (load file t)))
    (mapcar
     (lambda (l)
       (format t "%s\n" (car l))
       (align_pauses_utt (car l))
       t)
     p)
    t))

(define (align_pauses_utt name)
  "(align_utt name) 
Load in the actual aligned pause segments and put them into the synthesizer utterance."
  (let ((utt1 (Utterance Text nil))
	(silence (car (cadr (car (PhoneSet.description '(silences))))))
	actual-segments)
    
    (Initialize utt1)
    (utt.relation.load utt1 'Segment 
		       (format nil "lab/%s.lab" name))
    (mapcar
     (lambda (seg)
       (cond
	((and (not (string-equal (item.name seg) silence))
	      (or (string-equal (item.name seg) "sil")
		  (string-equal (item.name seg) "sp") ;hkt small pause
		  (string-equal (item.name seg) "H#")
		  (string-equal (item.name seg) "h#")))
	 (item.set_feat seg "name" silence))
	((and (string-equal (item.name seg) "DB")
	      (item.next seg))
	 (item.set_feat (item.next seg) "cl_end" (item.feat seg "end"))
	 (item.delete seg)
	 (set! seg nil)))
       (if seg
	   (item.set_feat seg "start" "F:standard+unisyn_start")))
     (utt.relation.items utt1 'Segment))

    (utt.set_feat utt1 "fileid" name)
    (utt.save utt1 (format nil "utt/%s.utt" name))
    t))



(define (build_utts file lexicon)
  "(build_utts file lexicon) 
For each utterance in prompt file, 
* synthesize from text, or, if APML, from tokens
* merge aligned phones in  lab/*.lab  with predicted segments
* load pitchmarks from       pm/*.pm
* load f0 from               f0/*.f0 if present
* save to                   utt/*.utt
  and                       utt/*.lab."
  (setup_phoneset_and_lexicon lexicon)
  (let ((p (load file t)))
    (mapcar
     (lambda (l)
       (format t "%s\n" (car l))
       (cond
	((eq? (cadr l) 'apml)
	   (align_utt_apml (car l)))
	((eq? (cadr l) 'xxml)
	   (align_utt_singing_xxml (car l)))
	(t 
	 (align_utt_text (car l) (cadr l))))
       t)
     p)
    t))

(define (build_utts_slf file lexicon)
  "(build_utts_slf file lexicon) 
For each utterance in prompt file, 
build utterance for slf alignmnet and
save to utt/*.utt & utt/*.lab."
  (setup_phoneset_and_lexicon lexicon)
  (let ((p (load file t)))
    (mapcar
     (lambda (l)
       (format t "%s\n" (car l))
       (cond
	;((eq? (cadr l) 'apml)
	;   (align_utt_apml (car l)))
	;((eq? (cadr l) 'xxml)
	;   (align_utt_singing_xxml (car l)))
	(t 
	 (align_utt_slf_text (car l) (cadr l))))
       t)
     p)
    t))


(define (build_simple_utts file lexicon)
  "(build_simple_utts file lexicon) 
Builds simple utts with no segment times."
  (let (utt (p (load file t)))
    (setup_phoneset_and_lexicon lexicon)
    (mapcar
     (lambda (l)
       (format t "%s\n" (car l))
       (utt.save (utt.synth_toSegment_text (eval (list 'Utterance 'Text (cadr l)))) (format nil "utt_simple/%s.utt" (car l)))
       t)
     p)
    t))


(define (align_utt_slf_text name text)
  "(align_utts_text file) 
Synth an utterance and load in the actualed aligned segments and merge
them into the synthesizer utterance."
  (let ((utt1 (utt.synth_toWord_text (eval (list 'Utterance 'Text text)))))
    (remove_punc_words utt1)
    (align_utt_slf_internal name utt1)
    (if(symbol-bound? 'expro)
       (begin
          (detect_quoting utt1 t) 
          (expro_emphasis utt1)))
  )
)


(define (align_utt_text name text)
  "(align_utts_text file) 
Synth an utterance and load in the actualed aligned segments and merge
them into the synthesizer utterance."
  (let ((utt1 (utt.synth_toSegment_text (eval (list 'Utterance 'Text text)))))
    (align_utt_internal name utt1)))

(define (align_utt_apml name)
  "(align_utts file) 
Synth an utterance and load in the actualed aligned segments and merge
them into the synthesizer utterance."
  (let ((utt1 (Utterance Tokens nil)))
    (utt.load utt1 (format nil "apml/%s.apml" name))
    (utt.synth_toSegment_tokens utt1)
    (align_utt_internal name utt1)))

;; synth to segments and save as mlf.
(define (align_utt_singing_xxml name)
  "(align_utt_singing_xxml name)
Synthesize given xxml and align labels"
  (let ((old_tts_hooks tts_hooks))
    (set! tts_hooks (list utt.synth_toSegment_tokens xxml_align_utt_wrapper))
    (set! msb_name name)
    (tts (format nil "xml/%s.xml" name) 'singing)
    (set! tts_hooks old_tts_hooks)))

(define (xxml_align_utt_wrapper utt)
  (singing_xxml_transfer_info utt)
  (align_utt_internal msb_name utt))

(define (singing_xxml_transfer_info utt)
  (let (f d)
  (mapcar
   (lambda (i)
     (set! f (item.feat i "freq"))
     (set! d (item.feat i "dur"))
     (if (atom f)
	 (item.set_feat i "freq" f) 
	 (item.set_feat i "freq" (car f)))
     (if (atom d)
	 (item.set_feat i "dur" d) 
	 (item.set_feat i "dur" (car d))))
     (utt.relation.items utt 'Token))))

(define (align_utt_slf_internal name utt1)
  "(align_utt_internal name utterace) 
Load in the aligned segments and merge them into the synthesized utterance."
  (let ((silence (car (cadr (car (PhoneSet.description '(silences))))))
	words segments (syl nil) (word nil) (seg nil) pms)

    (set! pms (track.load (format nil "pm/%s.pm" name)))
	
    (utt.relation.load utt1 'Segment
		       (format nil "lab/%s.lab" name))

    (set! words (utt.relation.items utt1 'Word))
    (set! segments (utt.relation.items utt1 'Segment))
    (utt.relation.create utt1 'Syllable)
    (utt.relation.create utt1 'SylStructure)


      (while (and segments (or word words))
        (set! seg (car segments))
	(item.set_feat seg "start" "F:standard+unisyn_start")
	(set! segments (cdr segments))
	;(format t "word is: %l   seg is %l\n" (if word (item.name word) nil) (item.name seg))
       (cond
	((string-matches (item.name seg) closure_regex) ; closures
	 (item.set_feat (item.next seg) 
			"cl_end" 
			(item.feat seg "end"))
	 (item.delete seg))
	((string-equal (item.name seg) ".") ; syl boundary
	 (item.delete seg)
	 (set! syl nil))
	((string-equal (item.name seg) "sp") ; sp/word boundary
	 (if (< (- (parse-number (item.feat seg "end"))
		   (parse-number (item.feat (item.prev seg) "end")))
		sp_threshold)
	     (begin
	       (item.delete seg)
	       (set! syl nil))
	     (item.set_name seg silence))
	  (set! word nil)) 
	((or (string-equal (item.name seg) "sil")
	     (string-equal (item.name seg) "#")
	     (string-equal (item.name seg) "H#")
	     (string-equal (item.name seg) "h#"))
	 (item.set_name seg silence))
	((or (string-equal (item.name seg) "start") ; ignore
	     (string-equal (item.name seg) "end"))
	 (item.delete seg))
	(t ; a actual segment or a stress mark
	 (if (null word)
	     (begin
	       (while (null word)
		  (set! word (car words))	       
		  (set! words (cdr words))
		  (if (string-equal (item.feat word "pos") "punc")
		      (set! word nil)))
	       (utt.relation.append utt1 'SylStructure word)))
	 (if (null syl)
	     (begin
	       (set! syl (utt.relation.append utt1 'Syllable "syl"))
	       (item.relation.append_daughter word 'SylStructure syl)
	       (item.set_feat syl 'stress 0)))
	 (cond
	  ((string-equal (item.name seg) "#1") ; stress mark
	   (item.set_feat syl 'stress 1)
	   (item.delete seg))
	  ((string-equal (item.name seg) "#2") ; stress mark
	   (item.set_feat syl 'stress 2)
	   (item.delete seg))
	  (t
	   (item.relation.append_daughter syl 'SylStructure seg)
	   (item.set_feat seg "name" (make_unsafe_name seg))
	   (if (not (check_pms seg pms))
	       (begin
		 (format t "Bad pitchmarking: %s %s.\n" (item.feat seg "name") (item.feat seg "end"))
		 (item.set_feat seg "bad" "no_pms")))))))) ; segment 

      (utt.synth_toWord_post_text utt1)
      (utt.mark_outoflex_words utt1)
      (utt.set_feat utt1 "fileid" name)
      ;; If we have an F0 add in targets too
      (if (probe_file (format nil "f0/%s.f0" name))
	  (build::add_targets utt1))
      (utt.save utt1 (format nil "utt/%s.utt" name))
      (utt.save.segs utt1 (format nil "utt/%s.lab" name))
      t))


(define (align_utt_internal name utt1)
  "(align_utt_internal name utterace) 
Load in the aligned segments and merge them into the synthesized utterance."
  ;(set! debug t)
  (let ((silence (car (cadr (car (PhoneSet.description '(silences))))))
	segments actual-segments pms)

    (set! pms (track.load (format nil "pm/%s.pm" name)))
	
    (utt.relation.load utt1 'actual-segment 
		       (format nil "lab/%s.lab" name))
    
    (set! segments (utt.relation.items utt1 'Segment))
    (set! actual-segments (utt.relation.items utt1 'actual-segment))

    ;(format t "       Segments are: %l\n" (mapcar (lambda (x) (item.name x)) segments))
    ;(format t "Actual Segments are: %l\n" (mapcar (lambda (x) (item.name x)) actual-segments))

    ;; undo 'safe segment' name mappings which cause filename problems
    (mapcar (lambda (seg) (item.set_feat seg "name" (make_unsafe_name seg))) actual-segments)

    ;; These should align, but if the labels had to be hand edited
    ;; then they may not. 
    (while (and segments actual-segments)

      (set! wname (item.feat (car segments) 'R:SylStructure.parent.parent.name))
      (if (string-equal wname "0")
         (set! wname (item.feat (car segments) 'n.R:SylStructure.parent.parent.name)))
      (if (and(string-equal (item.name (car actual-segments)) "#")
              (not(string-equal (item.feat (car actual-segments) 'n.name) "sil")))
        (format debug "\t\t\t\t\t%l:  \n" wname))

      (format debug "seg:%l    actual: %l\n" (item.name (car segments)) (item.name (car actual-segments)))
      (cond
       ;; PROCESS STOP AND AFFRICATE CLOSURES IN ACTUAL-SEGMENTS.
       ;; Delete `_cl` segments, from actual marking the end as the join point in the next token.
       ((string-matches (item.name (car actual-segments)) closure_regex)
	(item.set_feat (cadr actual-segments) 
		       "cl_end" 
		       (item.feat (car actual-segments) "end"))
	(set! actual-segments (cdr actual-segments)))
       ;; FIX WIERD SILENCES NAMES IN ACTUAL-SEGMENTS
       ;; Change all possible silence names to that specified by the first silence defined by the phoneset.
       ((and (not (string-equal (item.name (car actual-segments)) silence))
	     (or (string-equal (item.name (car actual-segments)) "sil")
		 (string-equal (item.name (car actual-segments)) "sp") ;hkt small pause
		 (string-equal (item.name (car actual-segments)) "H#")
		 (string-equal (item.name (car actual-segments)) "h#")))
	;[format t " Fixing silence name.\n"]
	(item.set_feat (car actual-segments) "name" silence))
       ;; DELETE ZERO LENGTH SILENCES FROM ACTUAL SEGMENTS
       ((and (string-equal (item.name (car actual-segments)) silence)
	     (item.prev (car actual-segments))
	     (eq (- (parse-number (item.feat (car actual-segments) "end"))
		    (parse-number (item.feat (item.prev (car actual-segments)) "end")))
		 0))
       	;[format t " Deleting zero length silence.\n"]
	(item.delete (car actual-segments))
	(set! actual-segments (cdr actual-segments)))
       ;; INSERT MISSMATCHED SILENCES THAT APPEAR IN ACTUAL-SEGMENTS
       ;; If a silence occurs in actual-segments then insert it in
       ;; segments if it is not too short.
       ((and (not (string-equal (item.name (car segments))

				(item.name (car actual-segments))))
	     (string-equal (item.name (car actual-segments)) silence))
	;[format t " mismatched silence from actual segments.\n"]
	;; If the current silence segment is not one of the short segment exceptions,
	;; insert a matching silence in segments.
	;;           ;; a short amount of silence 
	(if (not (or (and (item.prev (car actual-segments))
			  (< (- (parse-number (item.feat (car actual-segments) "end"))
				(parse-number (item.feat (item.prev (car actual-segments)) "end")))
			     sp_threshold))
		     ;; or silence that probably should be part of a following closure.
		     (and (item.prev (car actual-segments))
			  (item.next (car actual-segments))
			  (string-matches (item.name (item.next (car actual-segments))) closure_regex)
			  (< (- (parse-number (item.feat (car actual-segments) "end"))
				(parse-number (item.feat (item.prev (car actual-segments)) "end")))
			     sp_closure_threshold))))

	    (begin
	      ;[format t "inserting silence from actual segments.\n"]
	      (item.insert (car segments)
			   (list silence (list (list "end" (item.feat (car actual-segments) "end"))
					       (list "score" (normalise_score (car actual-segments)))
					       (list "start" "F:standard+unisyn_start")))
			   'before)))
	(set! actual-segments (cdr actual-segments)))
       ;; INSERT STOPS THAT APPEAR IN ACTUAL-SEGMENTS
       ;; If a stop has been removed by postlex_stop_deletion but does occur in actual-segments then undo its deletion
       ((and (not (string-equal (item.name (car segments))
                                (item.name (car actual-segments))))
             (string-equal (item.feat (car segments) 'left_context)
                           (item.name (car actual-segments))))
          (format t "\n\n\n\t\tinserting %s from actual segments.\n\n\n" (item.feat (car segments) 'left_context))
          (item.insert (car segments)
                       (list (item.feat (car segments) 'left_context)
                             (list (list "end" (item.feat (car actual-segments) "end"))
                                   (list "score" (normalise_score (car actual-segments)))
                                   (list "start" "F:standard+unisyn_start")))
                       'before)
        (set! actual-segments (cdr actual-segments)))

       ;; DELETE MISSMATCHED SILENCES THAT OCCUR IN SEGMENTS
       ;; If there is no corresponding silence in actual-segments that it has to go.
       ((and (not (string-equal (item.name (car segments))
				(item.name (car actual-segments))))
	     (or(string-equal (item.name (car segments)) silence)
	        (string-equal (item.name (car segments)) "B_300") ))
	(item.delete (car segments))
	(set! segments (cdr segments)))
       ;; VOWEL REDUCTION SUBSTITUTION
       ((and (not (string-equal (item.name (car segments))
				(item.name (car actual-segments))))
	     (and (string-equal (item.feat (car segments) 'ph_vc) "+")
		  (string-equal (item.feat (car actual-segments) 'ph_vlng) "a")))
	(if (eq? (parse-number (item.feat (car segments) "R:SylStructure.parent.stress")) 0)    ; old version
;	[if [eq? [parse-number [item.feat [car segments] "reducable"]] 1]
;	[if t   ; lets try doing this anyway. 
	    (begin
	      ;[format t " %f vowel reduction: %s -> %s lex reducable: %s\n" 
		;      [item.feat [car actual-segments] 'end]
		 ;     [item.name [car segments]] 
		  ;    [item.name [car actual-segments]] 
		   ;   [item.feat [car segments] 'reducable]]
	      (item.set_feat (car segments) "name" (item.name (car actual-segments))))
	    (begin
	      ;[format t " %f vowel reduction blocked : [%s -> %s]   lex reducable: %s\n"
		;      [item.feat [car actual-segments] 'end]
		 ;     [item.name [car segments]] 
		  ;    [item.name [car actual-segments]] 
		   ;   [item.feat [car segments] 'reducable]]
	      (item.set_feat (car actual-segments) "name" (item.name (car segments))))))
       ;; MATCHING SEGMENT
       ;; Eventually segments can actually match!
       ((string-equal (item.name (car segments))
		      (item.name (car actual-segments)))
	(item.set_feat (car segments) "end" (item.feat (car actual-segments) "end"))
	(item.set_feat (car segments) "score" (normalise_score (car actual-segments)))
	(item.set_feat (car segments) "start" "F:standard+unisyn_start")
	; copy cl_end for stops
	(if (> (item.feat (car actual-segments) "cl_end") 0)
	    (item.set_feat (car segments) "cl_end" (item.feat (car actual-segments) "cl_end")))
	; add a feature for dipthongs
	(if (string-equal (item.feat (car segments) "ph_vlng") "d")
	    (item.set_feat (car segments) "dipth" 1))
	; check for at least one pitchmark
	(if (not (check_pms (car segments) pms))
	    (begin
	      (format t "Bad pitchmarking: %s %s.\n" (item.feat (car segments) "name") (item.feat (car segments) "end"))
	      (item.set_feat (car segments) "bad" "no_pms")))
	(set! segments (cdr segments))
	(set! actual-segments (cdr actual-segments)))
       ;; ALIGNMENT MISSMATCH ERROR
       (t
	(format stderr
		"align missmatch at %s (%f) %s (%f)\t\t%s\n"
		(item.name (car segments))
		(item.feat (car segments) "end")
		(item.name (car actual-segments))
		(item.feat (car actual-segments) "end")
                name)
	(error)))
      )

    (mapcar
     (lambda (seg)
      ;; shorten and split sliences
      (while (and (string-equal (item.name seg) silence)
		  (> (item.feat seg "segment_duration") 0.300))
;              [format t "splitting %s silence of %f at %f\n"
;		      [item.name seg]
;                      [item.feat seg "segment_duration"]
;                      [item.feat seg "end"]]
              (cond
               ((string-equal "h#" (item.feat seg "p.name"))
                (item.set_feat (item.prev seg) "end"
                               (+ 0.150 (item.feat seg "p.end"))))
               ((and (string-equal silence (item.feat seg "p.name"))
                     (string-equal silence (item.feat seg "p.p.name")))
                (item.set_feat (item.prev seg) "end"
                               (+ 0.150 (item.feat seg "p.end")))
                (item.set_feat (item.prev seg) "name" silence))
               (t
                (item.insert seg
                             (list silence
                                   (list 
                                    (list "end" (+ 0.150 
						   (item.feat seg "p.end")))
				    (list "score" (item.feat seg "score"))
				    (list "start" "F:standard+unisyn_start")))
                             'before)))))
     (utt.relation.items utt1 'Segment))

    (utt.relation.delete utt1 'actual-segment)
    (utt.mark_outoflex_words utt1)
    (utt.set_feat utt1 "fileid" name)
    ;; If we have an F0 add in targets too
    (if (probe_file (format nil "f0/%s.f0" name))
	(build::add_targets utt1))
    (utt.save utt1 (format nil "utt/%s.utt" name))
    (utt.save.segs utt1 (format nil "utt/%s.lab" name))
    t))
;;;;;;;;;;;;;; end of definition (align_utt_internal name utt1) ;;;;;;;;;;;;;;;;;


(define (build::add_targets utt)
)

(define (check_pms_2 seg pms)
  "(check_pms seg pms)
Check there is at least one pitchmark in this segment. Or synthesis will fail!"
)

(define (check_pms seg pms)
  "(check_pms seg pms)
Check there is at least one pitchmark in this segment. Or synthesis will fail!"
  (let (start end)
    (set! end (item.feat seg "end"))
    (if (item.prev seg)
	(set! start (item.feat seg "start"))
	(set! start 0))
    (if (equal? (track.index_below pms start) (track.index_below pms end))
	nil
	t)))

(define (normalise_score seg)
"(normalise_score score)
Normalise alignmnet score by length."
(let (start end)
  (set! end (item.feat seg "end"))
  (if (item.prev seg)
      (set! start (item.feat seg "start"))
      (set! start 0))
  (exp (/ (parse-number (item.feat seg "score")) (- end start)))))


(define (add_duration_info_utts file durfile)
  "(add_duration_info utts file durfile) 
Load average durations info and for each utterances in prompt file mark outliers."
  (let ((uttslist (load file t))
	(durslist (load durfile t))
	(utt (Utterance Text ""))
	mean sd log_seg_dur)
    (mapcar
     (lambda (li)
       (format t "%s\n" (car li))
       (utt.load utt (format nil "utt/%s.utt" (car li)))
       ;; process utterance
       (mapcar
	(lambda (seg)
	  (if (assoc_string (item.feat seg "name") durslist)
	      (begin
		(set! mean (car (cadr (assoc_string (item.feat seg "name") durslist))))
		(set! sd   (car (cdr (cadr (assoc_string (item.feat seg "name") durslist)))))
		(set! log_seg_dur (log (- (item.feat seg "end") (item.feat seg "start"))))
		;[format t "seg: %s dur: %f mean: %f sd: %f range: %f:%f\n" [item.feat seg "name"] [exp log_seg_dur] mean sd [exp [- mean [* 2 sd]]] [exp [+ mean [* 3 sd]]]]
		(if (or (> log_seg_dur (+ mean (* 3 sd)))
			(< log_seg_dur (- mean (* 2 sd))))
		    (begin
		      ;[format t "duration outlier\n"]
		      (item.set_feat seg "bad_dur" "outlier")))))
	  t)
	(utt.relation.items utt 'Segment))
       (utt.save utt (format nil "utt/%s.utt" (car li)))
       t)
     uttslist)
    t))

(define (dump_word_labs file utt_dir word_dir)
  "(dump_word_labs file utt_dir word_dir) 
Assumes utterances are already built."
  (let ((p (load file t))
	(utt (Utterance Text nil))
	wordfile)
    
    (mapcar
     (lambda (l)
       (format t "%s\n" (car l))
       (utt.load utt (format nil "%s/%s.utt" utt_dir (car l)))
       (set! wordfile (format nil "%s/%s.word" word_dir (car l)))
       (utt.save.words utt wordfile)
       t)
     p)
    t))

(define (after_synth_save_wrapper utt)
  (utt.save.wave utt "utt.wav")
  utt)

(define (remove_punc_words utt)
  (mapcar
   (lambda (w)
     (let ((pos (item.feat w "pos")))
       (if (or (string-equal "punc" pos)
	       (string-equal "fpunc" pos))
	   (let ((pbreak (item.feat w "pbreak"))
		 (wp (item.relation w 'Phrase)))
	     (if (and (string-matches pbreak "BB?")
		      (item.relation.prev w 'Word))
		 (item.set_feat
		  (item.relation.prev w 'Word) "pbreak" pbreak))
	     (item.relation.remove w 'Word)
	     ;; can't refer to w as we've just deleted it
	     (item.relation.remove wp 'Phrase)))))
   (utt.relation.items utt 'Word)))

